"""Index file of the Graphical User Interface shipped next to melon, called Melon UI.
In order to start the gui, please follow the standard Qt procedure:

from PySide6 import QtWidgets
from melongui.mainwindow import MainWindow

if __name__ == "__main__":
    app = QtWidgets.QApplication()
    window = MainWindow()
    window.buildUI()
    window.start()
    window.show()
    sys.exit(app.exec())
"""
